#![no_std]
#![no_main]

use cortex_m_rt::{entry, exception};
#[cfg(feature = "defmt")]
use defmt_rtt as _;
use embassy_boot_rp::*;
use embassy_rp::Peripheral;
use embassy_time::Duration;
use embassy_rp::gpio::{Input, Level, Output, Pull};
use embassy_rp::peripherals;

const FLASH_SIZE: usize = 2 * 1024 * 1024;

// The pause bits are on by default after reset but I think Watchdog::enable()
// disables them so let's enable them, again.
fn watchdog_pause_on_debug(_watchdog: peripherals::WATCHDOG) {
    // This turns off the watchdog - no good! :-/
    //embassy_rp::watchdog::Watchdog::new(watchdog_copy).pause_on_debug(true);

    use embassy_rp::pac;
    let watchdog = pac::WATCHDOG;
    //SAFETY: We know which values have been set by WatchdogFlash / Watchdog.
    unsafe {
        watchdog.ctrl().write(|w| {
            w.set_pause_dbg0(true);
            w.set_pause_dbg1(true);
            w.set_pause_jtag(true);
            w.set_enable(true);
        });
    }
}

#[entry]
fn main() -> ! {
    let p = embassy_rp::init(Default::default());

    // Uncomment this if you are debugging the bootloader with debugger/RTT attached,
    // as it prevents a hard fault when accessing flash 'too early' after boot.
    #[cfg(feature = "defmt")]
    if true {
        for _ in 0..10000000 {
            cortex_m::asm::nop();
        }
    }

    unsafe {
        // dirty trick to keep PROGRAM_INFO despite `--gc-sections` because we cannot add
        // KEEP(...) in that part of the predefined linker script
        let _ = core::ptr::read_volatile(PROGRAM_INFO.as_ptr());
    }

    let mut led_b = Output::new(p.PIN_4, Level::Low);
    let _led_g = Output::new(p.PIN_5, Level::Low);
    let mut led_r = Output::new(p.PIN_6, Level::Low);
    let _button_boot2 = Input::new(p.PIN_11, Pull::Up);

    led_r.set_high();

    // This is a bit problematic because it will remain active when probe-rs-cli is programming
    // the flash and it will interrupt the programming. However, we cannot make this any longer
    // than 8.4 seconds so there isn't much that we can do.
    // Manual workaround: Put rp2040 into bootloader mode via a real reset before using probe-rs-cli.
    // Workaround: Turn on "pause on debug". This seems to work well enough, for now. If not, use the
    //   manual workaround above.
    let watchdog_interval = Duration::from_secs(8);
    let watchdog_copy = unsafe { p.WATCHDOG.clone_unchecked() };

    let mut bl: BootLoader = BootLoader::default();  // taking partition addresses from memory.x
    let flash = WatchdogFlash::<FLASH_SIZE>::start(p.FLASH, p.WATCHDOG, watchdog_interval);
    watchdog_pause_on_debug(watchdog_copy);
    let mut flash = BootFlash::new(flash);
    let start = bl.prepare(&mut SingleFlashConfig::new(&mut flash));
    core::mem::drop(flash);

    led_r.set_low();
    led_b.set_high();

    unsafe { bl.load(start) }
}

#[no_mangle]
#[cfg_attr(target_os = "none", link_section = ".HardFault.user")]
unsafe extern "C" fn HardFault() {
    cortex_m::peripheral::SCB::sys_reset();
}

#[exception]
unsafe fn DefaultHandler(_: i16) -> ! {
    const SCB_ICSR: *const u32 = 0xE000_ED04 as *const u32;
    let irqn = core::ptr::read_volatile(SCB_ICSR) as u8 as i16 - 16;

    panic!("DefaultHandler #{:?}", irqn);
}

#[panic_handler]
fn panic(_info: &core::panic::PanicInfo) -> ! {
    cortex_m::asm::udf();
}


// Program information.
// This must be in the first 256 bytes after BOOT2 (including the end marker). The first thing in
// the predefined linker script is .Reset so let's add it to that.
const BINARY_INFO_MARKER_START: u32 = 0x7188ebf2;
const BINARY_INFO_MARKER_END: u32 = 0xe71aa390;
const PROGRAM_INFO_FLASH_OFFSET: u32 = 0x10087000 - PROGRAM_INFO_TOTAL_SIZE;  // last 4k of active partition
const PROGRAM_INFO_TOTAL_SIZE: u32 = 4096;
const PROGRAM_INFO_TABLE_SIZE: u32 = 128;  // one pointer (4 bytes) for each piece of data, all must be valid pointers
const PROGRAM_INFO_COPY_TABLE_SIZE: u32 = 128;  // 3*u32 for each entry, picotool stops after 10 entries
#[used]
#[link_section = ".Reset"]
pub static PROGRAM_INFO: [u32; 5] = {
    extern "C" {
        static __bootloader_active_program_info_start: u32;
        static __bootloader_active_program_info_end: u32;
    }
    [
        BINARY_INFO_MARKER_START,
        //&__bootloader_active_program_info_start as *const u32 as u32,
        //&__bootloader_active_program_info_end as *const u32 as u32,
        PROGRAM_INFO_FLASH_OFFSET,
        PROGRAM_INFO_FLASH_OFFSET + PROGRAM_INFO_TABLE_SIZE,
        PROGRAM_INFO_FLASH_OFFSET + PROGRAM_INFO_TOTAL_SIZE - PROGRAM_INFO_COPY_TABLE_SIZE,
        BINARY_INFO_MARKER_END
    ]
};
